/*####################################################################
 FILE: TCN75A.h
 VERSION: 1.0
 PURPOSE: Digital Temperature Sensor with TCN75A library for Arduino

 HISTORY:
 Mirko Prosseda - Original version (20/11/2012)
#######################################################################*/


#ifndef TCN75A_h
#define TCN75A_h

#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h"
#else
#include "WProgram.h"
#endif

#define TCN75ALIB_VERSION "1.0"

class TCN75A
{

private:
    float temperature;                     // temperature value
    uint8_t baseaddress, address;
    void select_register(uint8_t reg);              // point to the selected register
    void write_register(uint8_t reg, uint8_t byte); // write a byte to the selected register

public:
    void begin();                          // initialize I2C connection
    float read();                          // read temperature values from TCN75A sensor
    uint8_t set_resolution(uint8_t res);   // set the measurement resolution
    uint8_t set_address(uint8_t add);      // set the I2C sensor address (0 to 7)
    void shutdown();                       // put the sensor in shutdown mode
    void turnon();                         // resume the sensor from the shutdown mode
    void enable_oneshot();                 // enable one-shot mode
    uint8_t set_fault_queue(uint8_t f);    // set the fault queue number
    uint8_t set_alert_polarity(uint8_t p); // set polarity of Alarm signal
    uint8_t set_comp_int(uint8_t b);       // set Comp/Int bit
    void set_hysteresis(uint8_t h);        // set temperature hysteresis
    void set_templimit(uint8_t l);         // set temperature limit
    int read_register(uint8_t reg, uint8_t bytes); // read an internal register
};
#endif
